#include "mainwindow.h"
#include "abakuscommon.h"

#include <kaccel.h>
#include <kmenubar.h>
#include <kaction.h>
#include <kstdaction.h>
#include <kshortcut.h>
#include <klistview.h>
#include <kdebug.h>
#include <kconfig.h>
#include <kglobal.h>
#include <kconfigbase.h>
#include <kactionclasses.h>

#include <qlayout.h>
#include <qradiobutton.h>
#include <qbuttongroup.h>
#include <qsplitter.h>

#include "function.h"
#include "error_handler.h"
#include "resultlistbox.h"
#include "resultlistboxtext.h"
#include "valuemanager.h"
#include "node.h"

MainWindow::MainWindow() : KMainWindow(0), m_popup(0), m_insert(false)
{
    m_mainSplitter = new QSplitter(this);
    QWidget *box = new QWidget(m_mainSplitter);
    QVBoxLayout *layout = new QVBoxLayout(box);
    m_layout = layout;
    layout->setSpacing(6);
    layout->setMargin(6);

    setCentralWidget(m_mainSplitter);

    QWidget *configBox = new QWidget(box);
    layout->addWidget(configBox);

    QHBoxLayout *configLayout = new QHBoxLayout(configBox);
    configLayout->addStretch(1);

    QButtonGroup *buttonGroup = new QButtonGroup(0);

    m_degrees = new QRadioButton(i18n("&Degrees"), configBox);
    configLayout->addWidget(m_degrees);
    buttonGroup->insert(m_degrees);
    slotDegrees();
    connect(m_degrees, SIGNAL(clicked()), SLOT(slotDegrees()));

    m_radians = new QRadioButton(i18n("&Radians"), configBox);
    configLayout->addWidget(m_radians);
    buttonGroup->insert(m_radians);
    connect(m_radians, SIGNAL(clicked()), SLOT(slotRadians()));

    m_history = new QVBox(box);
    layout->addWidget(m_history);
    m_history->setSpacing(6);
    m_history->setMargin(0);

    QLabel *label = new QLabel(i18n("History: "), m_history);
    label->setAlignment(AlignCenter);

    m_result = new ResultListBox(m_history);
    m_result->installEventFilter(this);
    m_result->setSelectionMode(QListBox::NoSelection);
    m_result->setHScrollBarMode(QListBox::AlwaysOff);
    connect(m_result, SIGNAL(signalEntrySelected(const QString &)),
                      SLOT(slotEntrySelected(const QString &)));
    connect(m_result, SIGNAL(signalResultSelected(const QString &)),
                      SLOT(slotResultSelected(const QString &)));

    m_history->setStretchFactor(m_result, 1);
    layout->setStretchFactor(m_history, 1);

    QHBox *editBox = new QHBox(box);
    layout->addWidget(editBox);
    editBox->setSpacing(6);

    m_edit = new KComboBox(editBox);
    m_edit->setFocus();
    m_edit->setEditable(true);
    m_edit->setDuplicatesEnabled(false);
    editBox->setStretchFactor(m_edit, 1);

    KPushButton *evalButton = new KPushButton(i18n("&Evaluate"), editBox);

    connect(evalButton, SIGNAL(clicked()), SLOT(slotEvaluate()));

    connect(m_edit, SIGNAL(returnPressed(const QString &)), SLOT(slotReturnPressed(const QString &)));
    connect(m_edit, SIGNAL(textChanged(const QString &)), SLOT(slotTextChanged(const QString &)));

    m_listSplitter = new QSplitter(Vertical, m_mainSplitter);
    m_fnList = new KListView(m_listSplitter);
    m_fnList->addColumn("Functions");
    m_fnList->addColumn("Value");
    m_fnList->setResizeMode(KListView::LastColumn);

    m_varList = new KListView(m_listSplitter);
    m_varList->addColumn("Variables");
    m_varList->addColumn("Value");
    m_varList->setResizeMode(KListView::LastColumn);

    connect(FunctionManager::instance(), SIGNAL(signalFunctionAdded(const QString &)),
            this, SLOT(slotNewFunction(const QString &)));
    connect(FunctionManager::instance(), SIGNAL(signalFunctionRemoved(const QString &)),
            this, SLOT(slotRemoveFunction(const QString &)));

    connect(ValueManager::instance(), SIGNAL(signalValueAdded(const QString &, number_t)),
            this, SLOT(slotNewValue(const QString &, number_t)));
    connect(ValueManager::instance(), SIGNAL(signalValueChanged(const QString &, number_t)),
            this, SLOT(slotChangeValue(const QString &, number_t)));
    connect(ValueManager::instance(), SIGNAL(signalValueRemoved(const QString &)),
            this, SLOT(slotRemoveValue(const QString &)));

    setupLayout();
    setupGUI(Keys | StatusBar | Save | Create);
}

bool MainWindow::event(QEvent *e)
{
    ErrorEvent *errorEvent = dynamic_cast<ErrorEvent *>(e);
    if(!errorEvent)
	return KMainWindow::event(e);

    m_lastError = errorEvent->message();
    return true;
}

bool MainWindow::eventFilter(QObject *o, QEvent *e)
{
    if(!o || !e)
	return false;

    if(o == m_result && dynamic_cast<QResizeEvent *>(e))
	m_result->triggerUpdate(true);

    return false;
}

bool MainWindow::queryExit()
{
    saveConfig();
    return KMainWindow::queryExit();
}

void MainWindow::contextMenuEvent(QContextMenuEvent *e)
{
    static KPopupMenu *popup = 0;

    if(!popup) {
	popup = new KPopupMenu(this);
	action("options_show_menubar")->plug(popup);
    }

    if(!action<KToggleAction>("options_show_menubar")->isChecked())
	popup->popup(e->globalPos());
}

void MainWindow::polish()
{
    KMainWindow::polish();
    loadConfig();
}

void MainWindow::slotReturnPressed(const QString &text)
{
    QString str = text;

    /* Add right parentheses as needed to balance out the expression. */
    int parenLevel = getParenthesesLevel(str);
    for(int i = 0; i < parenLevel; ++i)
	str += ')';
    
    m_lastError = QString::null;
    number_t result = parseString(str.latin1());

    QString resultVal;
    ResultListBoxText *after = 0;
    QListBoxItem *item;

    if(m_result->count() > 0)
	after = static_cast<ResultListBoxText *>(m_result->item(m_result->count() - 1));

    switch(lastParseStatusResult()) {
	case 0:
	    resultVal = QString::number(result, 'g', 18);

	    ValueManager::instance()->setValue("ans", result);
	    m_insert = true;
	break;

	case 2: // OK, no result to speak of
	    resultVal = "OK";
	    m_insert = false;
	break;

	default:
	    // We have an ErrorEvent pending, process it first.
	    qApp->processEvents();

	    resultVal = m_lastError;
	    m_insert = false;
    }

    item = new ResultListBoxText(m_result, str, resultVal, after, !m_insert);

    if(m_history->isShown())
	m_edit->setCurrentText(text);
    else
	m_edit->setEditText(resultVal);

    m_result->setCurrentItem(item);
    m_result->ensureCurrentVisible();

    QTimer::singleShot(0, m_edit->lineEdit(), SLOT(selectAll()));
}

void MainWindow::slotTextChanged(const QString &str)
{
    if(str.length() == 1 && m_insert) {
	QRegExp re("^[-+/*]");
	m_insert = false;
	if(str.find(re) != -1)
	    m_edit->lineEdit()->setText("ans " + str + " ");
    }
}

void MainWindow::slotEvaluate()
{
    slotReturnPressed(m_edit->currentText());
}

void MainWindow::slotUpdateSize()
{
    if(m_newSize != QSize(0, 0))
	resize(m_newSize);
    else
	resize(width(), minimumSize().height());
}

void MainWindow::slotDegrees()
{
    setTrigMode(Abakus::Degrees);
    m_degrees->setChecked(true);
    if(action("setDegreesMode"))
	action<KToggleAction>("setDegreesMode")->setChecked(true);
}

void MainWindow::slotRadians()
{
    setTrigMode(Abakus::Radians);
    m_radians->setChecked(true);
    if(action("setRadiansMode"))
	action<KToggleAction>("setRadiansMode")->setChecked(true);
}

int MainWindow::getParenthesesLevel(const QString &str)
{
    int level = 0;

    for(unsigned i = 0; i < str.length(); ++i)
	if(str[i] == '(')
	    ++level;
	else if(str[i] == ')')
	    --level;

    return level;
}

void MainWindow::loadConfig()
{
    {
	KConfigGroup config(KGlobal::config(), "Settings");

	QString mode = config.readEntry("Trigonometric mode", "Degrees");
	if(mode == "Degrees") {
	    setTrigMode(Abakus::Degrees);
	    m_degrees->setChecked(true);
	}
	else {
	    setTrigMode(Abakus::Radians);
	    m_radians->setChecked(true);
	}
    }

    {
	KConfigGroup config(KGlobal::config(), "Variables");

	QStringList list = config.readListEntry("Saved Variables");
	for(QStringList::ConstIterator it = list.begin(); it != list.end(); ++it) {
	    QStringList values = QStringList::split('=', *it);
	    if(values.count() != 2) {
		kdWarning() << "Your configuration file has somehow been corrupted!\n";
		continue;
	    }

	    ValueManager::instance()->setValue(values[0], values[1].toDouble());
	}
    }

    {
	KConfigGroup config(KGlobal::config(), "GUI");

	bool showHistory = config.readBoolEntry("ShowHistory", true);
	action<KToggleAction>("toggleHistoryList")->setChecked(showHistory);
	m_history->setShown(showHistory);

	bool showFunctions = config.readBoolEntry("ShowFunctions", true);
	action<KToggleAction>("toggleFunctionList")->setChecked(showFunctions);
	m_fnList->setShown(showFunctions);

	bool showVariables = config.readBoolEntry("ShowVariables", true);
	action<KToggleAction>("toggleVariableList")->setChecked(showVariables);
	m_varList->setShown(showVariables);
    }

    {
	KConfigGroup config(KGlobal::config(), "Functions");

	QStringList fnList = config.readListEntry("FunctionList");
	for(QStringList::ConstIterator it = fnList.begin(); it != fnList.end(); ++it)
	    parseString(*it); // Run the function definitions through the parser
    }

    populateListViews();
}

void MainWindow::saveConfig()
{
    {
	KConfigGroup config(KGlobal::config(), "Settings");

	if(trigMode() == Abakus::Degrees)
	    config.writeEntry("Trigonometric mode", "Degrees");
	else
	    config.writeEntry("Trigonometric mode", "Radians");
    }

    {
	KConfigGroup config(KGlobal::config(), "Variables");

	QStringList list;
	QStringList values = ValueManager::instance()->valueNames();
	QStringList::ConstIterator it = values.begin();

	for(; it != values.end(); ++it) {
	    if(ValueManager::instance()->isValueReadOnly(*it))
		continue;

	    list += QString("%1=%2")
	                .arg(*it)
		        .arg(ValueManager::instance()->value(*it), 0, 'g', 18);
	}

	config.writeEntry("Saved Variables", list);
    }

    {
	KConfigGroup config(KGlobal::config(), "GUI");

	config.writeEntry("ShowHistory", m_result->isShown());
	config.writeEntry("ShowFunctions", m_fnList->isShown());
	config.writeEntry("ShowVariables", m_varList->isShown());
    }

    {
	KConfigGroup config(KGlobal::config(), "Functions");

	FunctionManager *manager = FunctionManager::instance();

	QStringList userFunctions = manager->functionList(FunctionManager::UserDefined);
	QStringList saveList;

	for(QStringList::ConstIterator it = userFunctions.begin(); it != userFunctions.end(); ++it) {
	    UnaryFunction *fn = dynamic_cast<UnaryFunction *>(manager->function(*it)->userFn->fn);
	    QString var = manager->function(*it)->userFn->varName;
	    QString expr = fn->operand()->infixString();

	    saveList += QString("set %1(%2) = %3").arg(*it).arg(var).arg(expr);
	}

	config.writeEntry("FunctionList", saveList);
    }
}

void MainWindow::setupLayout()
{
    KActionCollection *ac = actionCollection();

    KStdAction::quit(kapp, SLOT(quit()), ac);
    KStdAction::showMenubar(this, SLOT(slotToggleMenuBar()), ac);

    KToggleAction *ta = new KToggleAction(i18n("&Degrees"), ALT + Key_D, this, SLOT(slotDegrees()), ac, "setDegreesMode");
    ta->setExclusiveGroup("TrigMode");
    ta->setChecked(trigMode() == Abakus::Degrees);

    ta = new KToggleAction(i18n("&Radians"), ALT + Key_R, this, SLOT(slotRadians()), ac, "setRadiansMode");
    ta->setExclusiveGroup("TrigMode");
    ta->setChecked(trigMode() == Abakus::Radians);

    ta = new KToggleAction(i18n("Show &History List"), ALT + Key_H, this, SLOT(slotToggleHistoryList()), ac, "toggleHistoryList");
    ta->setChecked(true);

    ta = new KToggleAction(i18n("Show &Variables"), ALT + Key_V, this, SLOT(slotToggleVariableList()), ac, "toggleVariableList");
    ta->setChecked(true);

    ta = new KToggleAction(i18n("Show &Functions"), ALT + Key_F, this, SLOT(slotToggleFunctionList()), ac, "toggleFunctionList");
    ta->setChecked(true);

    ta = new KToggleAction(i18n("Activate &Compact Mode"), ALT + Key_C, this, SLOT(slotToggleCompactMode()), ac, "toggleCompactMode");
    ta->setChecked(false);

    new KAction(i18n("Clear &History"), ALT + Key_L, m_result, SLOT(clear()), ac, "clearHistory");
}

void MainWindow::populateListViews()
{
    QStringList values = ValueManager::instance()->valueNames();

    number_t value = ValueManager::instance()->value("pi");
    new KListViewItem(m_varList, "pi", QString::number(value, 'g', 8));

    value = ValueManager::instance()->value("e");
    new KListViewItem(m_varList, "e", QString::number(value, 'g', 8));
}

KAction *MainWindow::action(const char *key)
{
    return actionCollection()->action(key);
}

void MainWindow::slotEntrySelected(const QString &text)
{
    m_edit->setEditText(text);
}

void MainWindow::slotResultSelected(const QString &text)
{
    m_edit->lineEdit()->insert(text);
}

void MainWindow::slotToggleMenuBar()
{
    menuBar()->setShown(!menuBar()->isShown());
}

void MainWindow::slotToggleFunctionList()
{
    bool show = action<KToggleAction>("toggleFunctionList")->isChecked();
    m_fnList->setShown(show);
}

void MainWindow::slotToggleVariableList()
{
    bool show = action<KToggleAction>("toggleVariableList")->isChecked();
    m_varList->setShown(show);
}

void MainWindow::slotToggleHistoryList()
{
    bool show = action<KToggleAction>("toggleHistoryList")->isChecked();
    m_history->setShown(show);
}

void MainWindow::slotNewFunction(const QString &name)
{
    UserFunction *userFn = FunctionManager::instance()->function(name)->userFn;
    UnaryFunction *fn = dynamic_cast<UnaryFunction *>(userFn->fn);
    QString fnName = QString("%1(%2)").arg(name, userFn->varName);
    QString expr = fn->operand()->infixString();

    new KListViewItem(m_fnList, fnName, expr);
}

void MainWindow::slotRemoveFunction(const QString &name)
{
    UserFunction *userFn = FunctionManager::instance()->function(name)->userFn;
    QString fnName = QString("%1(%2)").arg(name, userFn->varName);

    QListViewItem *item = 0;
    while((item = m_fnList->findItem(fnName, 0)) != 0)
	delete item;
}

void MainWindow::slotNewValue(const QString &name, number_t value)
{
    QString valueStr = QString::number(value, 'g', 8);
    new KListViewItem(m_varList, name, valueStr);
}

void MainWindow::slotChangeValue(const QString &name, number_t value)
{
    QListViewItem *item = m_varList->findItem(name, 0);
    if(item)
	item->setText(1, QString::number(value, 'g', 8));
}

void MainWindow::slotRemoveValue(const QString &name)
{
    delete m_varList->findItem(name, 0);
}

void MainWindow::slotToggleCompactMode()
{
    if(action<KToggleAction>("toggleCompactMode")->isChecked()) {
	m_wasFnShown = m_fnList->isShown();
	m_wasVarShown = m_varList->isShown();
	m_wasHistoryShown = m_history->isShown();

	m_fnList->setShown(false);
	m_varList->setShown(false);
	m_history->setShown(false);

	m_oldSize = size();
	m_newSize = QSize(0, 0);
	QTimer::singleShot(0, this, SLOT(slotUpdateSize()));
    }
    else {
	m_fnList->setShown(m_wasFnShown);
	m_varList->setShown(m_wasVarShown);
	m_history->setShown(m_wasHistoryShown);

	m_newSize = m_oldSize;
	QTimer::singleShot(0, this, SLOT(slotUpdateSize()));
    }
}

#include "mainwindow.moc"
