#include <kdebug.h>

#include <math.h>

#include "node.h"
#include "valuemanager.h"

void Node::deleteNode(Node *node)
{
    if(dynamic_cast<BaseFunction *>(node) != 0)
        delete node;
}

BaseFunction::BaseFunction(const char *name) :
    m_name(name), m_function(FunctionManager::instance()->function(name))
{
}

UnaryFunction::UnaryFunction(const char *name, Node *operand) :
    BaseFunction(name), m_node(operand)
{
}

UnaryFunction::~UnaryFunction()
{
    deleteNode(m_node);
    m_node = 0;
}

void UnaryFunction::setOperand(Node *operand)
{
    m_node = operand;
}

void UnaryFunction::applyMap(NodeFunctor &fn) const
{
    fn(operand());
    fn(this);
}

QString UnaryFunction::infixString() const
{
    return QString("%1(%2)").arg(name(), operand()->infixString());
}

UserDefinedFunction *UserDefinedFunction::clone() const
{
    Node *node = operand() ? operand()->clone() : 0;

    return new UserDefinedFunction(name(), node);
}

BuiltinFunction::BuiltinFunction(const char *name, Node *operand) :
    UnaryFunction(name, operand)
{
}

BuiltinFunction *BuiltinFunction::clone() const
{
    Node *node = operand() ? operand()->clone() : 0;

    return new BuiltinFunction(name(), node);
}

number_t BuiltinFunction::value() const
{
    if(function() && operand()) {
	number_t fnValue = operand()->value();
	return evaluateFunction(function(), fnValue);
    }

    return static_cast<number_t>(0);
}

UnaryOperator::UnaryOperator(Type type, Node *operand)
    : m_type(type), m_node(operand)
{
}

UnaryOperator::~UnaryOperator()
{
    deleteNode(m_node);
    m_node = 0;
}

UnaryOperator *UnaryOperator::clone() const
{
    Node *node = operand() ? operand()->clone() : 0;
    return new UnaryOperator(type(), node);
}

void UnaryOperator::applyMap(NodeFunctor &fn) const
{
    fn(operand());
    fn(this);
}

QString UnaryOperator::infixString() const
{
    if(dynamic_cast<BinaryOperator *>(operand()))
        return QString("-(%1)").arg(operand()->infixString());

    return QString("-%1").arg(operand()->infixString());
}

number_t UnaryOperator::value() const
{
    switch(type()) {
	case Negation:
	    return -(operand()->value());

	default:
	    kdError() << "Impossible case encountered for UnaryOperator!\n";
	    return static_cast<number_t>(0);
    }
}

BinaryOperator::BinaryOperator(Type type, Node *left, Node *right) :
    m_type(type), m_left(left), m_right(right)
{
}

BinaryOperator::~BinaryOperator()
{
    deleteNode(m_left);
    m_left = 0;

    deleteNode(m_right);
    m_right = 0;
}

BinaryOperator *BinaryOperator::clone() const
{
    Node *left = leftNode() ? leftNode()->clone() : 0;
    Node *right = rightNode() ? rightNode()->clone() : 0;

    return new BinaryOperator(type(), left, right);
}

void BinaryOperator::applyMap(NodeFunctor &fn) const
{
    fn(leftNode());
    fn(rightNode());
    fn(this);
}

QString BinaryOperator::infixString() const
{
    QString op;

    switch(type()) {
        case Addition:
            op = "+";
        break;

        case Subtraction:
            op = "-";
        break;

        case Multiplication:
            op = "*";
        break;

        case Division:
            op = "/";
        break;

        case Exponentiation:
            op = "^";
        break;

        default:
            op = "Error";
    }
    
    QString left = QString(isSimpleNode(leftNode()) ? "%1" : "(%1)").arg(leftNode()->infixString());
    QString right = QString(isSimpleNode(rightNode()) ? "%1" : "(%1)").arg(rightNode()->infixString());

    return QString("%1 %2 %3").arg(left, op, right);
}

number_t BinaryOperator::value() const
{
    if(!leftNode() || !rightNode()) {
	kdError() << "Can't evaluate binary operator!\n";
	return static_cast<number_t>(0);
    }

    number_t lValue = leftNode()->value();
    number_t rValue = rightNode()->value();

    switch(type()) {
	case Addition:
	    return lValue + rValue;

	case Subtraction:
	    return lValue - rValue;

	case Multiplication:
	    return lValue * rValue;

	case Division:
	    return lValue / rValue;

	case Exponentiation:
	    return powl(lValue, rValue);

	default:
	    kdError() << "Impossible case encountered evaluating binary operator!\n";
	    return static_cast<number_t>(0);
    }
}

bool BinaryOperator::isSimpleNode(Node *node) const
{
    if(dynamic_cast<Identifier *>(node) ||
       dynamic_cast<NumericValue *>(node) ||
       dynamic_cast<UnaryOperator *>(node) ||
       dynamic_cast<BaseFunction *>(node))
    {
        return true;
    }

    return false;
}

Identifier::Identifier(const char *name) : m_name(name)
{
}

number_t Identifier::value() const
{
    return ValueManager::instance()->value(name());
}

void Identifier::applyMap(NodeFunctor &fn) const
{
    fn(this);
}

// vim: set et ts=8 sw=4:
